@@ -47,6 +47,10 @@ module Data.Conduit.Combinators
     , sourceRandomN
     , sourceRandomGen
     , sourceRandomNGen
+    , sourceRandomWith
+    , sourceRandomNWith
+    , sourceRandomGenWith
+    , sourceRandomNGenWith
 
       -- ** Filesystem
     , sourceDirectory
@@ -141,6 +145,7 @@ module Data.Conduit.Combinators
     , mapWhile
     , conduitVector
     , scanl
+    , mapAccumWhile
     , concatMapAccum
     , intersperse
     , slidingWindow
@@ -162,6 +167,7 @@ module Data.Conduit.Combinators
     , filterME
     , iterM
     , scanlM
+    , mapAccumWhileM
     , concatMapAccumM
 
       -- ** Textual
@@ -179,6 +185,7 @@ module Data.Conduit.Combinators
 
       -- * Special
     , vectorBuilder
+    , mapAccumS
     , peekForever
     ) where
 
@@ -213,15 +220,12 @@ import           Data.Sequences.Lazy
 import qualified Data.Vector.Generic         as V
 import qualified Data.Vector.Generic.Mutable as VM
 import           Data.Void                   (absurd)
-import qualified Filesystem                  as F
-import           Filesystem.Path             (FilePath, (</>))
-import           Filesystem.Path.CurrentOS   (encodeString, decodeString)
 import           Prelude                     (Bool (..), Eq (..), Int,
-                                              Maybe (..), Monad (..), Num (..),
-                                              Ord (..), fromIntegral, maybe,
-                                              ($), Functor (..), Enum, seq, Show, Char, (||),
+                                              Maybe (..), Either (..), Monad (..), Num (..),
+                                              Ord (..), fromIntegral, maybe, either,
+                                              ($), Functor (..), Enum, seq, Show, Char,
                                               mod, otherwise, Either (..),
-                                              ($!), succ)
+                                              ($!), succ, FilePath)
 import Data.Word (Word8)
 import qualified Prelude
 import           System.IO                   (Handle)
@@ -234,14 +238,9 @@ import qualified System.Random.MWC as MWC
 import Data.Conduit.Combinators.Internal
 import Data.Conduit.Combinators.Stream
 import Data.Conduit.Internal.Fusion
-import qualified System.PosixCompat.Files as PosixC
 import           Data.Primitive.MutVar       (MutVar, newMutVar, readMutVar,
                                               writeMutVar)
 
-#ifndef WINDOWS
-import qualified System.Posix.Directory as Dir
-#endif
-
 -- Defines INLINE_RULE0, INLINE_RULE, STREAMING0, and STREAMING.
 #include "fusion-macros.h"
 
@@ -411,7 +410,7 @@ INLINE_RULE(replicateM, n m, CL.replicateM n m)
 --
 -- Since 1.0.0
 sourceFile :: (MonadResource m, IOData a) => FilePath -> Producer m a
-sourceFile fp = sourceIOHandle (F.openFile fp SIO.ReadMode)
+sourceFile fp = sourceIOHandle (SIO.openFile fp SIO.ReadMode)
 {-# INLINE sourceFile #-}
 
 -- | Read all data from the given @Handle@.
@@ -457,7 +456,8 @@ INLINE_RULE0(stdin, sourceHandle SIO.stdin)
 --
 -- Since 1.0.0
 sourceRandom :: (MWC.Variate a, MonadIO m) => Producer m a
-INLINE_RULE0(sourceRandom, initRepeat (liftIO MWC.createSystemRandom) (liftIO . MWC.uniform))
+sourceRandom = sourceRandomWith MWC.uniform
+{-# INLINE sourceRandom #-}
 
 -- | Create a stream of random values of length n, seeding from the system
 -- random number.
@@ -468,7 +468,8 @@ INLINE_RULE0(sourceRandom, initRepeat (liftIO MWC.createSystemRandom) (liftIO .
 sourceRandomN :: (MWC.Variate a, MonadIO m)
               => Int -- ^ count
               -> Producer m a
-INLINE_RULE(sourceRandomN, cnt, initReplicate (liftIO MWC.createSystemRandom) (liftIO . MWC.uniform) cnt)
+sourceRandomN cnt = sourceRandomNWith cnt MWC.uniform
+{-# INLINE sourceRandomN #-}
 
 -- | Create an infinite stream of random values, using the given random number
 -- generator.
@@ -479,7 +480,8 @@ INLINE_RULE(sourceRandomN, cnt, initReplicate (liftIO MWC.createSystemRandom) (l
 sourceRandomGen :: (MWC.Variate a, MonadBase base m, PrimMonad base)
                 => MWC.Gen (PrimState base)
                 -> Producer m a
-INLINE_RULE(sourceRandomGen, gen, initRepeat (return gen) (liftBase . MWC.uniform))
+sourceRandomGen gen = sourceRandomGenWith gen MWC.uniform
+{-# INLINE sourceRandomGen #-}
 
 -- | Create a stream of random values of length n, seeding from the system
 -- random number.
@@ -491,7 +493,54 @@ sourceRandomNGen :: (MWC.Variate a, MonadBase base m, PrimMonad base)
                  => MWC.Gen (PrimState base)
                  -> Int -- ^ count
                  -> Producer m a
-INLINE_RULE(sourceRandomNGen, gen cnt, initReplicate (return gen) (liftBase . MWC.uniform) cnt)
+sourceRandomNGen gen cnt = sourceRandomNGenWith gen cnt MWC.uniform
+{-# INLINE sourceRandomNGen #-}
+
+-- | Create an infinite stream of random values from an arbitrary distribution,
+-- seeding from the system random number.
+--
+-- Subject to fusion
+--
+-- Since 1.0.3
+sourceRandomWith :: (MWC.Variate a, MonadIO m) => (MWC.GenIO -> SIO.IO a) -> Producer m a
+INLINE_RULE(sourceRandomWith, f, initRepeat (liftIO MWC.createSystemRandom) (liftIO . f))
+
+-- | Create a stream of random values of length n from an arbitrary
+-- distribution, seeding from the system random number.
+--
+-- Subject to fusion
+--
+-- Since 1.0.3
+sourceRandomNWith :: (MWC.Variate a, MonadIO m)
+                  => Int -- ^ count
+                  -> (MWC.GenIO -> SIO.IO a)
+                  -> Producer m a
+INLINE_RULE(sourceRandomNWith, cnt f, initReplicate (liftIO MWC.createSystemRandom) (liftIO . f) cnt)
+
+-- | Create an infinite stream of random values from an arbitrary distribution,
+-- using the given random number generator.
+--
+-- Subject to fusion
+--
+-- Since 1.0.3
+sourceRandomGenWith :: (MWC.Variate a, MonadBase base m, PrimMonad base)
+                    => MWC.Gen (PrimState base)
+                    -> (MWC.Gen (PrimState base) -> base a)
+                    -> Producer m a
+INLINE_RULE(sourceRandomGenWith, gen f, initRepeat (return gen) (liftBase . f))
+
+-- | Create a stream of random values of length n from an arbitrary
+-- distribution, seeding from the system random number.
+--
+-- Subject to fusion
+--
+-- Since 1.0.3
+sourceRandomNGenWith :: (MWC.Variate a, MonadBase base m, PrimMonad base)
+                     => MWC.Gen (PrimState base)
+                     -> Int -- ^ count
+                     -> (MWC.Gen (PrimState base) -> base a)
+                     -> Producer m a
+INLINE_RULE(sourceRandomNGenWith, gen cnt f, initReplicate (return gen) (liftBase . f) cnt)
 
 -- | Stream the contents of the given directory, without traversing deeply.
 --
@@ -505,7 +554,7 @@ INLINE_RULE(sourceRandomNGen, gen cnt, initReplicate (return gen) (liftBase . MW
 --
 -- Since 1.0.0
 sourceDirectory :: MonadResource m => FilePath -> Producer m FilePath
-sourceDirectory = mapOutput decodeString . CF.sourceDirectory . encodeString
+sourceDirectory = CF.sourceDirectory
 
 -- | Deeply stream the contents of the given directory.
 --
@@ -518,7 +567,7 @@ sourceDirectoryDeep :: MonadResource m
                     => Bool -- ^ Follow directory symlinks
                     -> FilePath -- ^ Root directory
                     -> Producer m FilePath
-sourceDirectoryDeep follow = mapOutput decodeString . CF.sourceDirectoryDeep follow . encodeString
+sourceDirectoryDeep = CF.sourceDirectoryDeep
 
 -- | Ignore a certain number of values in the stream.
 --
@@ -541,11 +590,11 @@ dropE =
         then return ()
         else await >>= maybe (return ()) (go i)
 
-    go i seq = do
+    go i sq = do
         unless (onull y) $ leftover y
         loop i'
       where
-        (x, y) = Seq.splitAt i seq
+        (x, y) = Seq.splitAt i sq
         i' = i - fromIntegral (olength x)
 {-# INLINEABLE dropE #-}
 
@@ -573,10 +622,10 @@ dropWhileE f =
   where
     loop = await >>= maybe (return ()) go
 
-    go seq =
+    go sq =
         if onull x then loop else leftover x
       where
-        x = Seq.dropWhile f seq
+        x = Seq.dropWhile f sq
 {-# INLINE dropWhileE #-}
 
 -- | Monoidally combine all values in the stream.
@@ -803,7 +852,11 @@ INLINE_RULE(elem, x, any (== x))
 elemE :: (Monad m, Seq.EqSequence seq)
       => Element seq
       -> Consumer seq m Bool
+#if MIN_VERSION_mono_traversable(0,8,0)
+INLINE_RULE(elemE, f, any (oelem f))
+#else
 INLINE_RULE(elemE, f, any (Seq.elem f))
+#endif
 
 -- | Are no values in the stream equal to the given value?
 --
@@ -825,7 +878,11 @@ INLINE_RULE(notElem, x, all (/= x))
 notElemE :: (Monad m, Seq.EqSequence seq)
          => Element seq
          -> Consumer seq m Bool
+#if MIN_VERSION_mono_traversable(0,8,0)
+INLINE_RULE(notElemE, x, all (onotElem x))
+#else
 INLINE_RULE(notElemE, x, all (Seq.notElem x))
+#endif
 
 -- | Consume all incoming strict chunks into a lazy sequence.
 -- Note that the entirety of the sequence will be resident at memory.
@@ -1218,7 +1275,7 @@ INLINE_RULE(foldMapME, f, CL.foldM (ofoldlM (\accum e -> mappend accum `liftM` f
 --
 -- Since 1.0.0
 sinkFile :: (MonadResource m, IOData a) => FilePath -> Consumer a m ()
-sinkFile fp = sinkIOHandle (F.openFile fp SIO.WriteMode)
+sinkFile fp = sinkIOHandle (SIO.openFile fp SIO.WriteMode)
 {-# INLINE sinkFile #-}
 
 -- | Print all incoming values to stdout.
@@ -1350,12 +1407,12 @@ takeE =
         then return ()
         else await >>= maybe (return ()) (go i)
 
-    go i seq = do
+    go i sq = do
         unless (onull x) $ yield x
         unless (onull y) $ leftover y
         loop i'
       where
-        (x, y) = Seq.splitAt i seq
+        (x, y) = Seq.splitAt i sq
         i' = i - fromIntegral (olength x)
 {-# INLINEABLE takeE #-}
 
@@ -1389,13 +1446,13 @@ takeWhileE f =
   where
     loop = await >>= maybe (return ()) go
 
-    go seq = do
+    go sq = do
         unless (onull x) $ yield x
         if onull y
             then loop
             else leftover y
       where
-        (x, y) = Seq.span f seq
+        (x, y) = Seq.span f sq
 {-# INLINE takeWhileE #-}
 
 -- | Consume precisely the given number of values and feed them downstream.
@@ -1510,6 +1567,23 @@ scanlC f =
             loop seed'
 STREAMING(scanl, scanlC, scanlS, f x)
 
+-- | 'mapWhile' with a break condition dependent on a strict accumulator.
+-- Equivalently, 'CL.mapAccum' as long as the result is @Right@. Instead of
+-- producing a leftover, the breaking input determines the resulting
+-- accumulator via @Left@.
+--
+-- Subject to fusion
+mapAccumWhile, mapAccumWhileC :: Monad m =>
+    (a -> s -> Either s (s, b)) -> s -> ConduitM a b m s
+mapAccumWhileC f =
+    loop
+  where
+    loop !s = await >>= maybe (return s) go
+      where
+        go a = either (return $!) (\(s', b) -> yield b >> loop s') $ f a s
+{-# INLINE mapAccumWhileC #-}
+STREAMING(mapAccumWhile, mapAccumWhileC, mapAccumWhileS, f s)
+
 -- | 'concatMap' with an accumulator.
 --
 -- Subject to fusion
@@ -1752,6 +1826,20 @@ scanlMC f =
             loop seed'
 STREAMING(scanlM, scanlMC, scanlMS, f x)
 
+-- | Monadic `mapAccumWhile`.
+--
+-- Subject to fusion
+mapAccumWhileM, mapAccumWhileMC :: Monad m =>
+    (a -> s -> m (Either s (s, b))) -> s -> ConduitM a b m s
+mapAccumWhileMC f =
+    loop
+  where
+    loop !s = await >>= maybe (return s) go
+      where
+        go a = lift (f a s) >>= either (return $!) (\(s', b) -> yield b >> loop s')
+{-# INLINE mapAccumWhileMC #-}
+STREAMING(mapAccumWhileM, mapAccumWhileMC, mapAccumWhileMS, f s)
+
 -- | 'concatMapM' with an accumulator.
 --
 -- Subject to fusion
@@ -1993,6 +2081,20 @@ addE ref e = do
         else writeMutVar ref $! S idx' mv front
 {-# INLINE addE #-}
 
+-- | Consume a source with a strict accumulator, in a way piecewise defined by
+-- a controlling stream. The latter will be evaluated until it terminates.
+--
+-- >>> let f a s = liftM (:s) $ mapC (*a) =$ CL.take a
+-- >>> reverse $ runIdentity $ yieldMany [0..3] $$ mapAccumS f [] (yieldMany [1..])
+-- [[],[1],[4,6],[12,15,18]] :: [[Int]]
+mapAccumS :: Monad m => (a -> s -> Sink b m s) -> s -> Source m b -> Sink a m s
+mapAccumS f s xs = do
+    (zs, u) <- loop (newResumableSource xs, s)
+    lift (closeResumableSource zs) >> return u
+    where loop r@(ys, !t) = await >>= maybe (return r) go
+              where go a  = lift (ys $$++ f a t) >>= loop
+{-# INLINE mapAccumS #-}
+
 -- | Run a consuming conduit repeatedly, only stopping when there is no more
 -- data available from upstream.
 --
